#!/usr/bin/env python

import sys
import time

class GamePlayer:
	
	#GamePlayer init
	def __init__(self, activeLevel, startLevel, graceLevel):
		#Store level boundaries
		self.activeLevel = activeLevel
		self.startLevel = startLevel
		self.graceLevel = graceLevel
		self.reset()

	def reset(self):
		#Start at 0 level
		self.level = 0.0
		#Assume button is not pushed
		self.pushed = False
		#Start inactive
		self.active = False

	#Take a time step, increment or decrement depending on button pushed state
	def timeStep(self, time):
		if time <= 0.0:
			raise Exception('Invalid time step')
		if self.pushed:
			# Button is pushed - increment the level, but not past startLevel
			self.level = min( self.level + time, self.startLevel )
			#Set active on the way up
			if self.level >= self.activeLevel:
				self.active = True
		if not self.pushed:
			# Button not pushed
			# Drop to graceLevel, so letting go takes effect quicker
			if self.level > self.graceLevel:
				self.level = self.graceLevel

			# Decrement level, but don't go beyond zero
			self.level = max( self.level - time, 0.0 )
			#Unset active on the way down
			if self.level <= 0.0:
				self.active = False

	#Get current state
	@property
	def state(self):
		# Computed state is quite simple
		if self.start:
			return 'START'
		elif self.active:
			return 'ACTIVE'
		elif self.wait:
			return 'WAIT'
		return 'OUT'

	@property
	def start(self):
		return (self.level >= self.startLevel)

	@property
	def wait(self):
		return (self.level > 0.0)

class GameStarter:

	#Initialise game starter
	def __init__(self, maxPlayers, activeLevel, startLevel, graceLevel):
		#Raise error if number of players is too low
		if type(maxPlayers) != int:
			raise Exception('GameStarter.__init__: maxPlayers must be an integer greater than 2 (At least two players are required).')
		if maxPlayers < 2:
			raise Exception('GameStarter.__init__: maxPlayers must be an integer greater than 2 (At least two players are required). Attempted to init GameStarter with %d players.' % maxPlayers)
		#Raise error if startLevel or activeLevel is invalid
		if ((type(activeLevel) != float) or (type(startLevel) != float)):
			raise Exception('GameStarter.__init__: activeLevel must be a float greater than 0, startLevel must be a float greater than activeLevel.')
		if ((activeLevel <= 0.0) or (startLevel <= activeLevel)):
			raise Exception('GameStarter.__init__: activeLevel must be a float greater than 0, startLevel must be a float greater than activeLevel. (Active: %f, Start: %f)' % (activeLevel, startLevel))
		#Store maximum number of players
		self.maxPlayers = maxPlayers
		#Create this number of players
		self.players = [ GamePlayer(activeLevel, startLevel, graceLevel) for i in range(self.maxPlayers)]

	def resetAll(self):
		for pl in self.players:
			pl.reset()

	#get total number of players in given state
	def totalInState(self, state):
		return len(self.playersInState(state))

	#Check if a player will be active in this game
	def isStartablePlayer(self, player_id):
		return self.players[player_id].active

	def playersInState(self, state):
		return [id for id, pl in enumerate(self.players) if pl.state == state]

	#Get total number of startable players
	def totalStartablePlayers(self):
		return sum(1 for player in self.players if player.start)

	#Decide if a game is ready to start
	def shouldStart(self):
		#You should start if you have:
		# - at least two startable players
		# - at least one player who has reached the start state
		# - no players who have recently pressed (in)
		return (self.totalStartablePlayers() > 1) and (self.totalInState('START') > 0) and (self.totalInState('WAIT') == 0)

	
	#Push the given player's button
	def push(self, player_id):
		self.players[player_id].pushed = True

	#Release the given player's button
	def release(self, player_id):
		self.players[player_id].pushed = False

	#Check if the given player's button is pressed
	def isPushed(self, player_id):
		return self.players[player_id].pushed

	#Step all players by given time
	def timeStep(self, time):
		if (type(time) != float) or (time <= 0.0):
			raise Exception('GameStarter.timeStep: time step must be a positive float.')

		for i in range(self.maxPlayers):
			self.players[i].timeStep(time)

	#Get the level of the given player
	def getLevel(self, playerID):
		return self.players[playerID].level

	#Get the state of the given player
	def getState(self, playerID):
		return self.players[playerID].state

def main():

	#Visual test of GameStarter

	#Set level thresholds here
	activeLevel = 2.0
	startLevel = 5.0
	graceLevel = 1.0

	#Bar scale is number of characters that represent one second on the visualisation
	barScale = 20

	#Some maths for the time bar graphics
	activeBar = int(activeLevel * barScale)
	startBar = int(startLevel * barScale)
	activeBarString = '-' * (activeBar-1) + '|'
	startBarString = '-' * (startBar-activeBar-1) + '|'

	#Get an instance of GameStarter with four players
	starter = GameStarter(4, activeLevel, startLevel, graceLevel)

	#Print header for graphics
	print('ID|' + activeBarString + startBarString)
	#Pad lines ready for cursor moving back
	for i in range(4):
		print('')

	#Begin with players two and four pressed
	starter.push(1)
	starter.push(3)
	start = False
	totTime = 0.0;
	while (not start):

		#Set specific times for events to happen here
		if totTime > 3.0:
			starter.release(1)
		if totTime > 5.2:
			starter.push(0)
		if totTime > 6.0:
			starter.push(1)
		if totTime > 6.4:
			starter.push(2)
		#End of event timings

		#Do time calculations
		time.sleep(0.05)
		starter.timeStep(0.05)
		totTime += 0.05
		#Decide if game can start
		start = starter.shouldStart()
		
		#Step back four lines
		sys.stdout.write("\x1B[4A")
		#Print graphs
		for i in range(4):
			thislevel = starter.getLevel(i)
			print("%d |%s%s %s %s" % (i, ("#" * int(thislevel*barScale)), (" " * (int(barScale*startLevel)-int(barScale*thislevel))), starter.getState(i), str(starter.isPushed(i)) + '   '))

	#When game should start, get number of players in, print IDs
	numPlayersIn = 0

	print("Ready to start. Players:")
	for i in range(4):
		if starter.isStartablePlayer(i):
			numPlayersIn = numPlayersIn +1
			print("\tPlayer %d" % i)

	print("Start game with %d players." % numPlayersIn)

if __name__ == '__main__':
	main()
		









import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

    print(setuptools.find_packages())
setuptools.setup(
    name="game-starter",
    version="1.0.2",
    author="Daniel Bailey",
    author_email="danieljabailey@gmail.com",
    description="Decides when to start a game, based on who's holding a button.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/danieljabailey/GameStarter",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python

import unittest
from GameStarter import GameStarter, GamePlayer, main

class TestStartButtons(unittest.TestCase):
	def setUp(self):
		pass

	def tearDown(self):
		pass

	def test_player_init(self):
		#Ensure that the player can be instantiated and does sensible things
		#Note that this class should only be instantiated by the GameStarter and so has less validity check on its inputs
		pl = GamePlayer(1.0, 2.0, 0.5)
		self.assertEqual(0.0, pl.level)
		self.assertEqual("OUT", pl.state)
		self.assertEqual(False, pl.pushed)

	def test_player_reset(self):
		#Ensure that the player can be reset
		#Note that this class should only be instantiated by the GameStarter and so has less validity check on its inputs
		pl = GamePlayer(1.0, 2.0, 0.5)
		pl.pushed = True
		pl.timeStep(5.0)
		pl.reset()
		self.assertEqual(0.0, pl.level)
		self.assertEqual("OUT", pl.state)
		self.assertEqual(False, pl.pushed)

	def test_player_timing(self):
		#Ensure that the player can be instantiated and does sensible things
		#Note that this class should only be instantiated by the GameStarter and so has less validity check on its inputs
		pl = GamePlayer(1.0, 2.0, 0.5)
		pl.timeStep(0.5)
		self.assertEqual("OUT", pl.state)
		self.assertEqual(False, pl.pushed)
		self.assertEqual(0.0, pl.level)

		pl.pushed = True
		self.assertEqual(True, pl.pushed)

		pl.timeStep(0.51)
		self.assertEqual("WAIT", pl.state)
		level = pl.level
		self.assertTrue((0.509<level) and (level < 0.511))

		pl.timeStep(0.5)
		self.assertEqual("ACTIVE", pl.state)
		level = pl.level
		self.assertTrue((1.009<level) and (level < 1.011))

		pl.timeStep(1.0)
		self.assertEqual("START", pl.state)
		level = pl.level
		self.assertTrue((1.99<level) and (level < 2.01))

	def test_player_invalid_time_step(self):
		pl = GamePlayer(1.0, 2.0, 0.5)
		invalidTimes = [0, 0.0, -1, -1.0]
		for invalidTime in invalidTimes:
			pl.pushed = True
			self.assertRaises(Exception, pl.timeStep, invalidTime)
			pl.pushed = False
			self.assertRaises(Exception, pl.timeStep, invalidTime)

	def test_invalid_players(self):
		#It is invalid to have only one player. This should raise an exception. As should a negative number.
		invalidInputs = [1, 0, -1, None, False, True, "2", "foo"]
		for invalidInput in invalidInputs:
			try:
				gs = GameStarter(invalidInput, 1.0, 2.0, 0.5)
				self.fail('Started with 1 player')
			except Exception as e:
				errorMsg = 'GameStarter.__init__: maxPlayers must be an integer greater than 2 (At least two players are required).'
				self.assertEqual(errorMsg, str(e)[0:len(errorMsg)])

	def test_invalid_levels(self):
		#It is invalid to have a negative active level threshold. This should raise an exception, as should invalid types.
		invalidInputs = [(-1.0, 2.0, 0.5), (1.0, 0.5, 0.2), (1.0, -23.0, 0.5), (2, 1, 3), (1, 0, 4), (-3, -1, -4), (None, None, None), (True, True, True), (False, True, False), ("1", "2", "7")]
		for invalidInput in invalidInputs:
			try:
				gs = GameStarter(2, *invalidInput)
				self.fail('Started with invalid levels')
			except Exception as e:
				errorMsg = 'GameStarter.__init__: activeLevel must be a float greater than 0, startLevel must be a float greater than activeLevel.'
				self.assertEqual(errorMsg, str(e)[0:len(errorMsg)])

	def test_invalid_time_step(self):
		#It is invalid to have a negative time step. This should raise an exception
		invalidInputs = [-1, -1.0, -99999, "foo", "-23", "34", "12.34", None, True, False]
		for invalidInput in invalidInputs:
			try:
				gs = GameStarter(2, 1.0, 2.0, 0.5)
				gs.timeStep(invalidInput)
				self.fail('Invalid time step allowed')
			except Exception as e:
				errorMsg = 'GameStarter.timeStep: time step must be a positive float.'
				self.assertEqual(errorMsg, str(e)[0:len(errorMsg)])

	def test_total_state_zero(self):
		#totalInState should never raise an exception, it should always be 0 for an invalid input
		testInputs = ["foo", 1, 2, None, True, False]
		for testInput in testInputs:
			gs = GameStarter(2, 1.0, 2.0, 0.5)
			self.assertEqual(0, gs.totalInState(testInput))

	def test_total_state_correct(self):
		testInputs = [2, 3, 4, 5, 6, 10, 999]
		for testInput in testInputs:
			gs = GameStarter(testInput, 1.0, 2.0, 0.5)
			self.assertEqual(testInput, gs.totalInState("OUT"))

	def test_player_state_correct(self):
		testInputs = [2, 3, 4, 5, 6, 10, 999]
		for testInput in testInputs:
			gs = GameStarter(testInput, 1.0, 2.0, 0.5)
			self.assertEqual(list(range(testInput)), sorted(gs.playersInState("OUT")))

	def test_total_state_postreset(self):
		testInputs = [2, 3, 4, 5, 6, 10, 999]
		for testInput in testInputs:
			gs = GameStarter(testInput, 1.0, 2.0, 0.5)
			gs.push(1)
			gs.timeStep(1.5)
			self.assertEqual(1, gs.totalInState("ACTIVE"))
			self.assertEqual(testInput-1, gs.totalInState("OUT"))
			gs.resetAll()
			self.assertEqual(testInput, gs.totalInState("OUT"))

	def test_two_player_start(self):
		#test that two players can start a game
		try:
			gs = GameStarter(2, 1.0, 2.0, 0.5)
		except Exception as e:
			self.fail('Exception during __init__')
		#Both players push
		gs.push(0)
		gs.push(1)
		#Wait for three seconds
		gs.timeStep(3.0)
		#By now, shouldStart must be true and there should be 2 startable players
		self.assertTrue(gs.shouldStart())
		self.assertEqual(2, gs.totalStartablePlayers())

	def test_single_player_cant_start(self):
		#test that single players cannot start a game
		try:
			gs = GameStarter(2, 1.0, 2.0, 0.5)
		except Exception as e:
			self.fail('Exception during __init__')
		#One player pushes
		gs.push(0)
		#Wait for three seconds
		gs.timeStep(3.0)
		#shouldStart must be false
		self.assertFalse(gs.shouldStart())

	def test_player_can_drop_out(self):
		#test that a player can go below active and drop out of the round start
		try:
			gs = GameStarter(2, 1.0, 2.0, 0.5)
		except Exception as e:
			self.fail('Exception during __init__')
		#One player pushes
		gs.push(0)
		#Wait for three seconds
		gs.timeStep(3.0)
		#player has peaked at 2 seconds, let go of the button
		gs.release(0)
		#1.5 seconds later and they should already be out
		gs.timeStep(1.5)
		self.assertEqual("OUT", gs.getState(0))
		self.assertEqual(0.0, gs.getLevel(0))


	def test_two_player_start_with_four_player_game(self):
		#test that two players can start a game
		try:
			gs = GameStarter(4, 1.0, 2.0, 0.5)
		except Exception as e:
			self.fail('Exception during __init__')
		#Both players push
		gs.push(0)
		gs.push(1)
		#Wait for two seconds
		gs.timeStep(2.0)
		#By now, shouldStart must be true and there should be 2 startable players
		self.assertTrue(gs.shouldStart())
		self.assertEqual(2, gs.totalStartablePlayers())

	def test_button_spam_filtering(self):
		#test that two players can start a game even when someone is button spamming
		try:
			gs = GameStarter(4, 1.0, 2.0, 0.5)
		except Exception as e:
			self.fail('Exception during __init__')
		#Both players push
		gs.push(0)
		gs.push(1)
		#Wait for 1.5 seconds (nearly there)
		gs.timeStep(1.5)
		#player three then spams
		gs.push(2)
		gs.timeStep(0.7)
		gs.release(2)
		gs.timeStep(0.5)
		#By now, shouldStart must be true and there should be 2 startable players
		self.assertTrue(gs.shouldStart())
		self.assertEqual(2, gs.totalStartablePlayers())
		#Do more buttom mashing to see if the filter still works after this
		gs.push(2)
		gs.timeStep(0.7)
		gs.release(2)
		gs.timeStep(0.5)
		gs.push(2)
		gs.timeStep(0.7)
		gs.release(2)
		gs.timeStep(0.5)
		#shouldStart must stil be true and there should still be 2 startable players
		self.assertTrue(gs.shouldStart())
		self.assertEqual(2, gs.totalStartablePlayers())

	def test_late_joiners(self):
		#test that for players can start a game even when some people take a while to join in
		try:
			gs = GameStarter(4, 1.0, 2.0, 0.5)
		except Exception as e:
			self.fail('Exception during __init__')
		#One players pushes
		gs.push(0)
		#Wait for 3 seconds (player 1 fully ready)...
		gs.timeStep(3.0)
		#Player two joins in
		gs.push(1)
		#a little later...
		gs.timeStep(0.8)
		#player three then joins
		gs.push(2)
		#later still..
		gs.timeStep(0.7)
		#player four pushes too
		gs.push(3)
		#another two seconds...
		gs.timeStep(2.0)
		#and now, shouldStart must be true and there should be 4 startable players
		self.assertTrue(gs.shouldStart())
		self.assertEqual(4, gs.totalStartablePlayers())

	def test_dodgy_button(self):
		#test that a dodgy button that flickers on and off sometimes doesn't cause problems
		#this also simulates people who fail to keep their hand on the button persistently
		try:
			gs = GameStarter(2, 1.0, 20.0, 0.5) #note 20 second start time on this one
		except Exception as e:
			self.fail('Exception during __init__')
		#Both players push
		gs.push(0)
		gs.push(1)
		#Wait for 1.5 seconds (nearly there)...
		gs.timeStep(1.5)
		#Player two goes dodgy
		gs.release(1)
		gs.timeStep(0.1)
		gs.push(1)
		gs.timeStep(0.3)
		gs.release(1)
		gs.timeStep(0.2)
		gs.push(1)
		gs.timeStep(0.7)
		gs.release(1)
		gs.timeStep(0.04)
		gs.push(1)
		gs.timeStep(10.8)
		gs.release(1)
		gs.timeStep(0.2)
		gs.push(1)
		gs.timeStep(0.7)
		gs.release(1)
		gs.timeStep(0.04)
		gs.push(1)
		gs.timeStep(20.0) #long one to make sure

		#shouldStart must be true and there should be 2 startable players
		self.assertTrue(gs.shouldStart())
		self.assertEqual(2, gs.totalStartablePlayers())
		
	def test_main_run(self):
		main()
		

if __name__ == '__main__':
    unittest.main()
